﻿#include "cuda_runtime.h"
#include "device_launch_parameters.h"
#include "cudnn.h"
#include <iostream>

using namespace std;

// Macros for error checking
#define checkCUDNN(expression)                               \
  {                                                          \
    cudnnStatus_t status = (expression);                     \
    if (status != CUDNN_STATUS_SUCCESS) {                    \
      std::cout << "Error on line " << __LINE__ << ": "      \
                << cudnnGetErrorString(status) << std::endl; \
      throw new exception();                                 \
    }                                                        \
  }


int main()
{
	// Loop many times
	for (int i = 0; i < 100000; i++)
	{
		cudnnHandle_t cudnn(0);									// general cudnn handle
		cudnnFilterDescriptor_t filter_weights_descriptor(0);	// filter weights descriptor
		cudnnTensorDescriptor_t filter_bias_descriptor(0);		// bias size descriptor
		cudnnConvolutionDescriptor_t convolution_descriptor(0); // convolution descriptor
		cudnnTensorDescriptor_t input_tensor_descriptor(0);		// input data descriptor
		cudnnActivationDescriptor_t  activation_descriptor(0);	// activation descriptor

		size_t memory_input_size;	// input memory
		float* cuda_memory_input(0);

		size_t memory_output_size;	// output memory
		float* cuda_memory_output(0);

		size_t memory_weights_size;	// convolution weights
		float* cuda_memory_weights(0);

		size_t memory_bias_size;	// convolution bias
		float* cuda_memory_bias(0);

		size_t memory_workspace_size;	// workspace
		void* cuda_memory_workspace(0);

		// Constants
		// Number of input and output layers = 1
		int h = 10, w = 10; // layer size
		int conv_size = 3;  // convolution size


		checkCUDNN(cudnnCreate(&cudnn));
		cudaDeviceSynchronize();

		memory_weights_size = conv_size * conv_size * sizeof(float);
		memory_bias_size = sizeof(float);
		memory_input_size = h * w * sizeof(float);
		memory_output_size = h * w * sizeof(float);

		cudaMalloc(&cuda_memory_input, memory_input_size);
		cudaMalloc(&cuda_memory_output, memory_output_size);
		cudaMalloc(&cuda_memory_weights, memory_weights_size);
		cudaMalloc(&cuda_memory_bias, memory_bias_size);

		checkCUDNN(cudnnCreateFilterDescriptor(&filter_weights_descriptor));
		checkCUDNN(cudnnSetFilter4dDescriptor(filter_weights_descriptor,
			/*dataType=*/CUDNN_DATA_FLOAT,
			/*format=*/CUDNN_TENSOR_NCHW,
			/*out_channels=*/1,
			/*in_channels=*/1,
			/*kernel_height=*/conv_size,
			/*kernel_width=*/conv_size));

		checkCUDNN(cudnnCreateTensorDescriptor(&filter_bias_descriptor));
		checkCUDNN(cudnnSetTensor4dDescriptor(filter_bias_descriptor,
			/*format=*/CUDNN_TENSOR_NCHW,
			/*dataType=*/CUDNN_DATA_FLOAT,
			/*batch_size=*/1,
			/*channels=*/1,
			/*image_height=*/1,
			/*image_width=*/1));

		checkCUDNN(cudnnCreateConvolutionDescriptor(&convolution_descriptor));
		checkCUDNN(cudnnSetConvolution2dDescriptor(convolution_descriptor,
			/*pad_height=*/(conv_size - 1) / 2,
			/*pad_width=*/ (conv_size - 1) / 2,
			/*vertical_stride=*/1,
			/*horizontal_stride=*/1,
			/*dilation_height=*/1,
			/*dilation_width=*/1,
			/*mode=*/CUDNN_CROSS_CORRELATION,
			/*computeType=*/CUDNN_DATA_FLOAT));



		checkCUDNN(cudnnCreateActivationDescriptor(&activation_descriptor));
		checkCUDNN(cudnnSetActivationDescriptor(activation_descriptor,
			/*mode=*/CUDNN_ACTIVATION_RELU,
			/*reluNanOpt=*/CUDNN_NOT_PROPAGATE_NAN,
			/*coef=*/0.0));

		checkCUDNN(cudnnCreateTensorDescriptor(&input_tensor_descriptor));
		checkCUDNN(cudnnSetTensor4dDescriptor(input_tensor_descriptor,
			/*format=*/CUDNN_TENSOR_NCHW,
			/*dataType=*/CUDNN_DATA_FLOAT,
			/*batch_size=*/1,
			/*channels=*/1,
			/*image_height=*/h,
			/*image_width=*/w));


		checkCUDNN(cudnnGetConvolutionForwardWorkspaceSize(cudnn,
			input_tensor_descriptor,
			filter_weights_descriptor,
			convolution_descriptor,
			input_tensor_descriptor,
			CUDNN_CONVOLUTION_FWD_ALGO_IMPLICIT_PRECOMP_GEMM,
			&memory_workspace_size));

		if (memory_workspace_size > 0)
			cudaMalloc(&cuda_memory_workspace, memory_workspace_size);	// could be not allocated if workspace size == 0


		const float alpha1 = 1.0f, alpha2 = 0.0f;

		// THE FOLLOWING OPERATION PRODUCES MEMORY LEAK
		// WITHOUT IT THERE IS NO MEMORY LEAK
		
		checkCUDNN(cudnnConvolutionBiasActivationForward(cudnn,
			&alpha1,
			input_tensor_descriptor,
			cuda_memory_input,
			filter_weights_descriptor,
			cuda_memory_weights,
			convolution_descriptor,
			CUDNN_CONVOLUTION_FWD_ALGO_IMPLICIT_PRECOMP_GEMM,
			cuda_memory_workspace,
			memory_workspace_size,
			&alpha2,
			input_tensor_descriptor,
			cuda_memory_output,
			filter_bias_descriptor,
			cuda_memory_bias,
			activation_descriptor,
			input_tensor_descriptor,
			cuda_memory_output));
		
	
		// destroy all created descriptors
		cudnnDestroyActivationDescriptor(activation_descriptor);
		cudnnDestroyConvolutionDescriptor(convolution_descriptor);
		cudnnDestroyTensorDescriptor(filter_bias_descriptor);
		cudnnDestroyFilterDescriptor(filter_weights_descriptor);
		cudnnDestroyTensorDescriptor(input_tensor_descriptor);

		// free all allocated memory
		cudaFree(cuda_memory_input);
		cudaFree(cuda_memory_output);
		cudaFree(cuda_memory_weights);
		cudaFree(cuda_memory_bias);

		if (cuda_memory_workspace)
			cudaFree(cuda_memory_workspace);	// could be not allocated if workspace size == 0

		// finally, destroy cudnn handle
		checkCUDNN(cudnnDestroy(cudnn));
		cudaDeviceSynchronize();

		cout << "Iteration " << i << " " << cudaGetErrorString(cudaGetLastError()) << endl;
	}

	cout << "Press any key to continue..." << endl;
	getchar();

	return 0;
}